(* ProofTraceToStateBase contains definitions and proofs common to
   ProofTraceToStateInterferance, ProofTraceToStateStability,
   and ProofTraceToStateMain. *)



Require Import List.
Require Import SemanticsBase.
Require Import SemanticsState.
Require Import SemanticsTrace.
Require Import ProofPre.
Require Import ProofBase.



(* ~~~~~~~ LABELS_TO_PRE ~~~~~~~ *)

Fixpoint Labels_to_pre {a a' l p} (r : req (ll a a') l) (i : In p (down a)) (i' : In p (up a')) : In p (pre a') :=
  match a as a return req (ll a a') l -> In p (down a) -> In p (up a') -> In p (pre a') with
    | step (send finish A) =>
      fun r => fun i => fun i' =>
      match eq_Probe_dec p (bar A) with
        | right ne => match not_in_neq ne i with end
        | left e => (fix F {a'} := fun r : req (ll (step (send finish A)) a') l => fun i' =>
                     match a' as a' return req (ll _ a') l -> In (bar A) (up a') -> In p (pre a') with
                       | step (send start B) =>
                         fun r : req (ll (step (send finish A)) _) l => fun i' =>
                         match eq_Channel_dec A B with
                           | right ne' => match not_in_neq (proj1 (neq_Channel_to_Probe ne')) i' with end
                           | left e' => match (eq_ind_r (fun x => ~ In (hat x) _) (proj2 r) e') (proj1 r) with end end
                       | step (send use B) =>
                         fun r : req (ll (step (send finish A)) _) l => fun i' =>
                         match eq_Channel_dec A B with
                           | right ne' => match not_in_neq (proj1 (neq_Channel_to_Probe ne')) i' with end
                           | left e' => eq_ind_r (fun x => In x _) (eq_ind_r (fun x => In (bar x) _) (in_eq _ _) e') e end
                       | step (recv start B) => fun _ => fun i' => match not_in_neq (proj1 neq_Probe) i' with end
                       | step (recv use B) => fun _ => fun i' => match not_in_neq (proj1 neq_Probe) i' with end
                       | ll a0 a1 =>
                         fun r => fun i' =>
                         match in_app_or _ _ _ i' with
                           | or_introl il => in_or_app _ _ _ (or_introl (F (conj (proj1 r) (proj1 (proj2 r))) il))
                           | or_intror ir => in_or_app _ _ _ (or_intror (F (conj (proj1 r) (proj2 (proj2 r))) ir)) end
                       | _ => fun _ => fun i' => match in_nil i' with end
                     end r i') a' r (eq_ind_r (fun x => In x _) i' (eq_sym e)) end
    | step (recv finish A) =>
      fun r => fun i => fun i' =>
      match eq_Probe_dec p (hat A) with
        | right ne => match not_in_neq ne i with end
        | left e => (fix F {a'} := fun r : req (ll (step (recv finish A)) a') l => fun i' =>
                     match a' as a' return req (ll _ a') l -> In (hat A) (up a') -> In p (pre a') with
                       | step (recv start B) =>
                         fun r : req (ll (step (recv finish A)) _) l => fun i' =>
                         match eq_Channel_dec A B with
                           | right ne' => match not_in_neq (proj2 (neq_Channel_to_Probe ne')) i' with end
                           | left e' => match (proj1 r) (eq_ind_r (fun x => In (bar x) _) (proj2 r) e') with end end
                       | step (recv use B) =>
                         fun r : req (ll (step (recv finish A)) _) l => fun i' =>
                         match eq_Channel_dec A B with
                           | right ne' => match not_in_neq (proj2 (neq_Channel_to_Probe ne')) i' with end
                           | left e' => eq_ind_r (fun x => In x _) (eq_ind_r (fun x => In (hat x) _) (in_eq _ _) e') e end
                       | step (send start B) => fun _ => fun i' => match not_in_neq (proj2 neq_Probe) i' with end
                       | step (send use B) => fun _ => fun i' => match not_in_neq (proj2 neq_Probe) i' with end
                       | ll a0 a1 =>
                         fun r => fun i' =>
                         match in_app_or _ _ _ i' with
                           | or_introl il => in_or_app _ _ _ (or_introl (F (conj (proj1 r) (proj1 (proj2 r))) il))
                           | or_intror ir => in_or_app _ _ _ (or_intror (F (conj (proj1 r) (proj2 (proj2 r))) ir)) end
                       | _ => fun _ => fun i' => match in_nil i' with end
                     end r i') a' r (eq_ind_r (fun x => In x _) i' (eq_sym e)) end
    | ll a0 a1 => fun r => fun i => fun i' =>
                  match in_app_or _ _ _ i with
                    | or_introl il => Labels_to_pre (conj (proj1 (proj1 r)) (proj2 r)) il i'
                    | or_intror ir => Labels_to_pre (conj (proj2 (proj1 r)) (proj2 r)) ir i' end
    | _ => fun _ => fun i => fun _ => match in_nil i with end
  end r i i'.



(* ~~~~~~~ TRACEDOWNCOUNT_TO_ES ~~~~~~~ *)

Fixpoint TraceDownCount_to_ES' {P P' a l p} (t : Trace P a P') (r : req a l)
                               (i : In p (down a) /\ count_occ eq_Probe_dec l p > 1)
                                  : ErrorStep P l :=
  match t in Trace P a P' return req a l -> (In p (down a) /\ count_occ eq_Probe_dec l p > 1) -> ErrorStep P l with
    | CommB2_T A => fun r => fun i => match eq_Probe_dec p (bar A) with
                               | left e => CommB2_E _ _ r (eq_ind_r (fun x => count_occ _ _ x > 1) (proj2 i) (eq_sym e))
                               | right ne => match not_in_neq ne (proj1 i) with end end
    | CommQ2_T A => fun r => fun i => match eq_Probe_dec p (hat A) with
                               | left e => CommQ2_E _ _ r (eq_ind_r (fun x => count_occ _ _ x > 1) (proj2 i) (eq_sym e))
                               | right ne => match not_in_neq ne (proj1 i) with end end
    | Sequ_T _ _ q _ t' => fun r => fun i => Sequ_E _ _ _ (TraceDownCount_to_ES' t' r i)
    | Par_T _ _ _ _ a0 a1 t0 t1 => fun r => fun i => match i with conj id c =>
                                   match in_dec eq_Probe_dec p (down a0), in_dec eq_Probe_dec p (down a1) with
                                     | left id0, _ => ParL_E _ _ _ (TraceDownCount_to_ES' t0 (proj1 r) (conj id0 c))
                                     | _, left id1 => ParR_E _ _ _ (TraceDownCount_to_ES' t1 (proj2 r) (conj id1 c))
                                     | right nid0, right nid1 => match match in_app_or _ _ _ id with
                                                                         | or_introl id0 => nid0 id0
                                                                         | or_intror id1 => nid1 id1 end with end end end
    | _ => fun _ => fun i => match in_nil (proj1 i) with end
  end r i.

Definition TraceDownCount_to_ES {P P' a p} (t : Trace P a P') (c : compliant a)
                                (i : In p (down a) /\ count_occ eq_Probe_dec (pre a) p > 1)
                                   : ErrorStep P (pre a) :=
  TraceDownCount_to_ES' t c i.