(* ProofTraceToStateMain maps
   a compliant Traces to a BigStep
   or a BigStep and ErrorStep,
   defining:
    - Trace_to_BSorBSES *)



Require Import List.
Require Import SemanticsBase.
Require Import SemanticsState.
Require Import SemanticsTrace.
Require Import ProofPre.
Require Import ProofBase.
Require Import ProofTraceToStateBase.
Require Import ProofTraceToStateStability.



(* ~~~~~~~ COMPLIANT TRACES TO NON-ERRONEOUS OR ERRONEOUS STEP SEQUENCES ~~~~~~~ *)

Fixpoint Sequ_BS {P P' Q s s'} (bs : BigStep P s P' s')
                                   : BigStep (sequ P Q) s (sequ P' Q) s' :=
  match bs with
    | S' _ _ => S'
    | SS _ _ _ _ ss => SS (Sequ_S _ _ _ _ _ ss)
    | BS _ _ _ _ _ _ _ bs' se bs'' => BS (Sequ_BS bs') se (Sequ_BS bs'')
  end.

Fixpoint ParL_BS {P P' Q s s'} (bs : BigStep P s P' s')
                  : BigStep (par P Q) s (par P' Q) s' :=
  match bs with
    | S' _ _ => S'
    | SS _ _ _ _ ss => SS (ParL_S _ _ _ _ _ ss)
    | BS _ _ _ _ _ _ _ bs' se bs'' => BS (ParL_BS bs') se (ParL_BS bs'')
  end.

Fixpoint ParR_BS {P Q Q' s s'} (bs : BigStep Q s Q' s')
                  : BigStep (par P Q) s (par P Q') s' :=
  match bs with
    | S' _ _ => S'
    | SS _ _ _ _ ss => SS (ParR_S _ _ _ _ _ ss)
    | BS _ _ _ _ _ _ _ bs' se bs'' => BS (ParR_BS bs') se (ParR_BS bs'')
  end.

Fixpoint Trace_to_BSorBSES' {P P' a l} (t : Trace P a P') (r : req a l) (f' : forall p, In p (pre a) -> In p l)
                            (f : forall p, In p (down a) -> count_occ eq_Probe_dec l p = 1
                                                         /\ count_occ eq_Probe_dec (pre a) p = 1)
                               : sum (BigStep P l P' (Label_to_newState a l))
                                     {P'' : Program & {s'' : State & prod (BigStep P l P'' s'')
                                                                          (ErrorStep P'' s'')}} :=
  match t as t in Trace P a P' return req a l -> (forall p, In p (pre a) -> In p l)
  -> (forall p, In p (down a) -> count_occ eq_Probe_dec l p = 1
                              /\ count_occ eq_Probe_dec (pre a) p = 1)
  -> sum (BigStep P l P' (Label_to_newState a l))
         {P'' : Program & {s'' : State & prod (BigStep P l P'' s'')
                                              (ErrorStep P'' s'')}} with
    | CommB1_T _ => fun r => fun _ => fun _ => inl (SS (CommB1_S _ _ r))
    | CommB2_T _ => fun r => fun _ => fun f => inl (SS (CommB2_S _ _ r (proj1 (f _ (in_eq _ _)))))
    | CommQ1_T _ => fun r => fun _ => fun _ => inl (SS (CommQ1_S _ _ r))
    | CommQ2_T _ => fun r => fun _ => fun f => inl (SS (CommQ2_S _ _ r (proj1 (f _ (in_eq _ _)))))
    | SequSkip_T _ => fun _ => fun _ => fun _ => inl (SS (SequSkip_S _ _))
    | Sequ_T _ _ _ _ t' => fun r => fun f' => fun f =>
                           match Trace_to_BSorBSES' t' r f' f with
                             | inl bs => inl (Sequ_BS bs)
                             | inr (existT _ (existT _ (bs, es))) => inr (existT _ _ (existT _ _ (Sequ_BS bs, Sequ_E _ _ _ es))) end
    | ParSkip_T => fun _ => fun _ => fun _ => inl (SS (ParSkip_S _))
    | Par_T _ _ _ _ a0 a1 t0 t1 => fun r => fun f' => fun f =>
                                   match Trace_to_BSorBSES' t0 (proj1 r)
                                         (fun p => fun i => f' p (in_or_app _ _ _ (or_introl i)))
                                         (fun p => fun i => let X := (in_or_app _ _ _ (or_introl i))
                                                            in conj (proj1 (f p X))
                                                                    (proj2 (proj1 (in_count_occ_app_one (proj2 (f p X)))
                                                                                  (down_to_pre i)))) with
                                     | inr (existT _ (existT _ (bs, es))) => inr (existT _ _ (existT _ _ (ParL_BS bs, ParL_E _ _ _ es)))
                                     | inl bs =>
                                   match Trace_to_BSorBSES' t1 (proj2 r)
                                         (fun p => fun i => f' p (in_or_app _ _ _ (or_intror i)))
                                         (fun p => fun i => let X := (in_or_app _ _ _ (or_intror i))
                                                            in conj (proj1 (f p X))
                                                                    (proj2 (proj2 (in_count_occ_app_one (proj2 (f p X)))
                                                                                  (down_to_pre i)))) with
                                     | inr (existT _ (existT _ (bs', es'))) => inr (existT _ _ (existT _ _ (ParR_BS bs', ParR_E _ _ _ es')))
                                     | inl _ =>
                                   match req_dec a1 (Label_to_newState a0 l) with
                                     | right nr10 => match Trace_to_NSE t0 t1 (proj1 r) (proj2 r) nr10
                                                           (fun p => fun i => f' p (in_or_app _ _ _ (or_introl i)))
                                         (fun p => fun i => let X := (in_or_app _ _ _ (or_introl i))
                                                            in (conj (proj1 (f p X))
                                                               (proj2 (proj1 (in_count_occ_app_one (proj2 (f p X)))
                                                               (down_to_pre i))))) with
                                                     existT _ (existT _ (ss, ns)) => inr (existT _ _ (existT _ _ (S', ParL_NSE _ _ _ _ _ ss ns))) end
                                     | left r10 =>
                                   match req_dec a0 (Label_to_newState a1 l) with
                                     | right nr01 => match Trace_to_NSE t1 t0 (proj2 r) (proj1 r) nr01
                                                           (fun p => fun i => f' p (in_or_app _ _ _ (or_intror i)))
                                         (fun p => fun i => let X := (in_or_app _ _ _ (or_intror i))
                                                            in (conj (proj1 (f p X))
                                                               (proj2 (proj2 (in_count_occ_app_one (proj2 (f p X)))
                                                               (down_to_pre i))))) with
                                                     existT _ (existT _ (ss, ns)) => inr (existT _ _ (existT _ _ (S', ParR_NSE _ _ _ _ _ ss ns))) end
                                     | left r01 => match Trace_to_BSorBSES' t1 r10
                                                         (fun p => fun i => let X := (in_or_app _ _ _ (or_intror i))
                                                                            in Label_to_Down a0 (f' p X)
                                                         (fun ida0 => let Y := (in_or_app _ _ _ (or_introl ida0))
                                                                            in proj1 (proj2 (in_count_occ_app_one (proj2 (f p Y)))
                                                                            i) (down_to_pre ida0)))
                                                         (fun p => fun i => let X := (in_or_app _ _ _ (or_intror i)) in
                                                         (conj (eq_trans (Label_to_DownCount a0
                                                         (fun id0 => proj1 (proj1 (in_count_occ_app_one (proj2 (f p X)))
                                                         (down_to_pre id0)) (down_to_pre i))
                                                         (fun ip0 => proj1 (proj1 (in_count_occ_app_one (proj2 (f p X)))
                                                         (Labels_to_pre (conj (proj2 r) (proj1 r)) i ip0)) (down_to_pre i)))
                                                         (proj1 (f p X)))
                                                         (proj2 (proj2 (in_count_occ_app_one (proj2 (f p X)))
                                                                                             (down_to_pre i))))) with
                                                     | inl bs' => inl (BS (ParL_BS bs) eq_State_refl (ParR_BS (bs')))
                                                     | inr (existT _ (existT _ (bs', es'))) =>
                                                       inr (existT _ _ (existT _ _
                                                           (BS (ParL_BS bs) eq_State_refl (ParR_BS (bs')), ParR_E _ _ _ es'))) end end end end end
    | Rep_T _ => fun _ => fun _ => fun _ => inl (SS (Rep_S _ _))
    | DetL_T _ _ => fun r => fun _ => fun _ => inl (SS (DetL_S _ _ _ (proj1 (eval_s1 _) r)))
    | DetR_T _ _ => fun r => fun _ => fun _ => inl (SS (DetR_S _ _ _ (proj1 (eval_s1 _) r)))
    | NDetL_T _ _ => fun r => fun _ => fun _ => inl (SS (NDetL_S _ _ _ (proj1 (eval_s1 _) r)))
    | NDetR_T _ _ => fun r => fun _ => fun _ => inl (SS (NDetR_S _ _ _ (proj1 (eval_s1 _) r)))
    | _ => fun _ => fun _ => fun _ => inl S'
  end r f' f.

Fixpoint Trace_to_Pre {P P' a} (t : Trace P a P') : Program_to_State P = pre a :=
  match t as t in Trace P a P' return Program_to_State P = pre a with
    | Sequ_T _ _ _ _ t' => Trace_to_Pre t'
    | Par_T _ _ _ _ _ _ t0 t1 => f_equal2 _ (Trace_to_Pre t0) (Trace_to_Pre t1)
    | _ => eq_refl _
  end.

Fixpoint Trace_to_Post {P P' a} (t : Trace P a P') (v : valid P) : Program_to_State P' = post a :=
  match t as t in Trace P a P' return valid P -> Program_to_State P' = post a with
    | SequSkip_T _ => fun v => initial_to_nil (proj2 v)
    | Sequ_T _ _ _ _ t' => fun v => Trace_to_Post t' (proj1 v)
    | Par_T _ _ _ _ _ _ t0 t1 => fun v => f_equal2 _ (Trace_to_Post t0 (proj1 v)) (Trace_to_Post t1 (proj2 v))
    | Rep_T _ => fun v => initial_to_nil v
    | DetL_T (_, p0) _ => fun v => initial_to_nil (proj1 v)
    | DetR_T _ (_, p1) => fun v => initial_to_nil (proj2 v)
    | NDetL_T (_, p0) _ => fun v => initial_to_nil (proj1 v)
    | NDetR_T _ (_, p1) => fun v => initial_to_nil (proj2 v)
    | _ => fun _ => eq_refl _
  end v.

Definition Trace_to_BSorBSES {P P' a} (t : Trace P a P') (v : valid P) (c : compliant a)
                           : sum (BigStep P (Program_to_State P) P' (Program_to_State P'))
                                 {P'' : Program & {s'' : State & prod (BigStep P (Program_to_State P) P'' s'')
                                                                      (ErrorStep P'' s'')}} :=
  match Label_dec a with
    | inl f =>
  match Trace_to_BSorBSES' t c (fun p => fun i => i) (fun p => fun i => conj (f p i) (f p i)) with
    | inl bs => inl (BS (eq_rect_r (fun x => BigStep _ x _ _) bs (Trace_to_Pre t))
                (fun p => (eq_ind_r (fun x => eq_State _ x)
                                    (Label_to_PrePost a c f) (Trace_to_Post t v) p)) S')
    | inr bses => inr (eq_rect_r (fun x => {P'' : Program &
                                           {s'' : State & prod (BigStep _ x _ _) (_)}})
                                 bses (Trace_to_Pre t)) end
    | inr p => match p with exist _ ic => inr (existT _ _ (existT _ _
               (S', eq_rect_r (fun x => ErrorStep _ x) (TraceDownCount_to_ES t c ic) (Trace_to_Pre t)))) end
  end.