(* ProofTraceToStateStability maps
   nonstable Traces to ErrorSteps,
   defining:
    - TraceNotStable_to_ES *)



Require Import List.
Require Import SemanticsBase.
Require Import SemanticsState.
Require Import SemanticsTrace.
Require Import ProofPre.
Require Import ProofBase.
Require Import ProofTraceToStateBase.



(* ~~~~~~~ NON-STABLE TRACES TO ERRONEOUS STEPS ~~~~~~~ *)

Fixpoint Label_to_PostDown {p l} (a : Label) (i : In p (Label_to_newState a l))
                         : In p (post a) \/ (In p l /\ ~ In p (down a)) :=
  match a as a return In p (Label_to_newState a l) -> In p (post a) \/ (In p l /\ ~ In p (down a)) with
    | step (send start A) => fun i => match in_inv i with
                                          | or_introl e => or_introl (eq_ind_r (fun x => In x _) (in_eq _ _) (eq_sym e))
                                          | or_intror il => or_intror (conj il (fun x => in_nil x)) end
    | step (recv start A) => fun i => match in_inv i with
                                          | or_introl e => or_introl (eq_ind_r (fun x => In x _) (in_eq _ _) (eq_sym e))
                                          | or_intror il => or_intror (conj il (fun x => in_nil x)) end
    | step (send finish A) => fun i => match eq_Probe_dec p (bar A) with
                                           | left e => match remove_In _ _ _ (eq_ind_r (fun x => In x _) i (eq_sym e)) with end
                                           | right ne => or_intror (conj (in_remove_weak i) (not_in_neq ne)) end
    | step (recv finish A) => fun i => match eq_Probe_dec p (hat A) with
                                           | left e => match remove_In _ _ _ (eq_ind_r (fun x => In x _) i (eq_sym e)) with end
                                           | right ne => or_intror (conj (in_remove_weak i) (not_in_neq ne)) end
    | ll a0 a1 => fun i => match Label_to_PostDown a1 i with
                             | or_introl iu => or_introl (in_or_app _ _ _ (or_intror iu))
                             | or_intror il1 =>
                           match Label_to_PostDown a0 (proj1 il1) with
                             | or_introl iu => or_introl (in_or_app _ _ _ (or_introl iu))
                             | or_intror il0 => or_intror (conj (proj1 il0)
                                                (fun x => match in_app_or _ _ _ x with
                                                            | or_introl i0 => proj2 il0 i0
                                                            | or_intror i1 => proj2 il1 i1 end)) end end
    | _ => fun i => or_intror (conj i (fun x => in_nil x))
  end i.

Fixpoint Label_to_Down {p l} (a : Label) (i : In p l) (i' : ~ In p (down a)) : In p (Label_to_newState a l) :=
  match a as a return In p l -> ~ In p (down a) -> In p (Label_to_newState a l) with
    | step (send start A) => fun i => fun _ => in_cons _ _ _ i
    | step (recv start A) => fun i => fun _ => in_cons _ _ _ i
    | step (send finish A) => fun i => fun i' => match eq_Probe_dec p (bar A) with
                                                     | left e => match i' (eq_ind_r (fun x => In x _) (in_eq _ _) e) with end
                                                     | right ne => in_remove_neq ne i end
    | step (recv finish A) => fun i => fun i' => match eq_Probe_dec p (hat A) with
                                                     | left e => match i' (eq_ind_r (fun x => In x _) (in_eq _ _) e) with end
                                                     | right ne => in_remove_neq ne i end
    | ll a0 a1 => fun i => fun i' => Label_to_Down a1 (Label_to_Down a0 i (fun x => i' (in_or_app _ _ _ (or_introl x))))
                                                      (fun x => i' (in_or_app _ _ _ (or_intror x)))
    | _ => fun i => fun _ => i
  end i i'.

Fixpoint Label_to_Post {p s l} (a : Label) (r : req a s) (i : In p (post a)) (f : In p (pre a) -> In p l)
                               (f2 : In p (down a) -> count_occ eq_Probe_dec (pre a) p = 1) : In p (Label_to_newState a l) :=
    match a as a return req a s -> In p (post a) -> (In p (pre a) -> In p l)
    -> (In p (down a) -> count_occ eq_Probe_dec (pre a) p = 1) -> In p (Label_to_newState a l) with
    | step (send start A) => fun _ => fun i => fun _ => fun _ =>
                               match eq_Probe_dec p (bar A) with
                                 | left e => eq_ind_r (fun x => In x _) (in_eq _ _) e
                                 | right ne => match not_in_neq ne i with end end
    | step (recv start A) => fun _ => fun i => fun _ => fun _ =>
                               match eq_Probe_dec p (hat A) with
                                 | left e => eq_ind_r (fun x => In x _) (in_eq _ _) e
                                 | right ne => match not_in_neq ne i with end end
    | step (send use A) => fun _ => fun i => fun f => fun _ =>
                             match eq_Probe_dec p (bar A) with
                               | left e => f (eq_ind_r (fun x => In x _) (in_eq _ _) e)
                               | right ne => match not_in_neq ne i with end end
    | step (recv use A) => fun _ => fun i => fun f => fun _ =>
                             match eq_Probe_dec p (hat A) with
                               | left e => f (eq_ind_r (fun x => In x _) (in_eq _ _) e)
                               | right ne => match not_in_neq ne i with end end
    | ll a0 a1 => fun r => fun i => fun f => fun f2 =>
                  match in_app_or _ _ _ i with
                    | or_intror i1 => Label_to_Post a1 (proj2 r) i1
                                      (fun x => Label_to_Down a0 (f (in_or_app _ _ _ (or_intror x)))
                                       match in_dec eq_Probe_dec p (down a0) with
                                         | left id0 => match proj1 (proj1 (in_count_occ_app_one
                                                                          (f2 (in_or_app _ _ _ (or_introl id0))))
                                                                          (down_to_pre id0)) x with end
                                         | right nid0 => nid0 end)
                                      (fun x => proj2 (proj2 (in_count_occ_app_one
                                                             (f2 (in_or_app _ _ _ (or_intror x))))
                                                             (down_to_pre x)))
                    | or_introl i0 => match in_dec eq_Probe_dec p (down a1) with
                                        | left id1 => match proj1 (proj2 (in_count_occ_app_one
                                                                         (f2 (in_or_app _ _ _ (or_intror id1))))
                                                                         (down_to_pre id1))
                                                                         (Labels_to_pre (conj (proj2 r) (proj1 r)) id1 i0) with end
                                        | right nid1 =>
                                      Label_to_Down a1 (Label_to_Post a0 (proj1 r) i0
                                                       (fun x => f (in_or_app _ _ _ (or_introl x)))
                                                       (fun x => proj2 (proj1 (in_count_occ_app_one
                                                                              (f2 (in_or_app _ _ _ (or_introl x))))
                                                                              (down_to_pre x)))) nid1 end end
    | _ => fun _ => fun i => fun _ => fun _ => match in_nil i with end
  end r i f f2.

Fixpoint req_inv {a a' l l' s} (x : req a' s) (f : forall p, In p (pre a') -> In p l)
                               (f' : forall p, In p (pre a') -> In p l')
                               (f2 : forall p, In p (down a') -> count_occ eq_Probe_dec (pre a') p = 1)
                               (r : req a l) (r' : req a (Label_to_newState a' l))
                               (r'' : req a l')  : req a (Label_to_newState a' l') :=
  let F := fix ff g := match g as g return req (holds g) l
                       -> req (holds g) (Label_to_newState a' l)
                       -> req (holds g) l' -> req (holds g) (Label_to_newState a' l') with
                         | nil => fun _ => fun _ => fun _ => I
                         | A :: g' => fun il => fun i_l => fun il' => conj (
                                      match in_dec eq_Probe_dec (bar A) (post a') with
                                        | left ip => Label_to_Post a' x ip (f' _) (f2 _)
                                        | right nip => Label_to_Down a' (proj1 il') (fun id => match Label_to_PostDown a' (proj1 i_l) with
                                                                                                       | or_introl ip => nip ip
                                                                                                       | or_intror (conj _ nid) => nid id end) end)
                                                            (ff g' (proj2 il) (proj2 i_l) (proj2 il'))
                       end in
  match a as a return req a l -> req a (Label_to_newState a' l)
  -> req a l' -> req a (Label_to_newState a' l') with
    | step (send start A) => fun nil => fun ni_l => fun nil' =>
                               (fun i_l' => match Label_to_PostDown a' i_l' with
                                              | or_introl ip => ni_l (Label_to_Post a' x ip (f _) (f2 _))
                                              | or_intror (conj il' _) => nil' il' end)
    | step (recv start A) => fun il => fun i_l => fun il' =>
                               match in_dec eq_Probe_dec (bar A) (post a') with
                                 | left ip => Label_to_Post a' x ip (f' _) (f2 _)
                                 | right nip => Label_to_Down a' il' (fun id => match Label_to_PostDown a' i_l with
                                                                                  | or_introl ip => nip ip
                                                                                  | or_intror (conj _ nid) => nid id end) end
    | step (send finish A) => fun il => fun i_l => fun il' =>
                                match in_dec eq_Probe_dec (hat A) (post a') with
                                  | left ip => Label_to_Post a' x ip (f' _) (f2 _)
                                  | right nip => Label_to_Down a' il' (fun id => match Label_to_PostDown a' i_l with
                                                                                   | or_introl ip => nip ip
                                                                                   | or_intror (conj _ nid) => nid id end) end
    | step (recv finish A) => fun nil => fun ni_l => fun nil' =>
                               (fun i_l' => match Label_to_PostDown a' i_l' with
                                              | or_introl ip => ni_l (Label_to_Post a' x ip (f _) (f2 _))
                                              | or_intror (conj il' _) => nil' il' end)
    | ll a0 a1 => fun r => fun r' => fun r'' => conj (req_inv x f f' f2 (proj1 r) (proj1 r') (proj1 r''))
                                                     (req_inv x f f' f2 (proj2 r) (proj2 r') (proj2 r''))
    | holds G => fun r => fun r' => fun r'' => F G r r' r''
    | holds2 G1 G2 => fun r => fun r' => fun r'' => conj (F G1 (proj1 r) (proj1 r') (proj1 r'')) (F G2 (proj2 r) (proj2 r') (proj2 r''))
    | _ => fun r => fun r' => fun r'' => I
  end r r' r''.

Fixpoint req_dec (a : Label) (s : State) : {req a s} + {~ req a s} :=
  let F := fix f g := match g as g return {eval g s} + {~ eval g s} with
                        | nil => left I
                        | a :: g' => match in_dec eq_Probe_dec _ s, f g' with
                                       | left r, left r' => left (conj r r')
                                       | right nr, _ => right (fun r => nr (proj1 r))
                                       | _, right nr => right (fun r => nr (proj2 r)) end
                      end in
  match a as a return {req a s} + {~ req a s} with
    | step (send start A) => match in_dec eq_Probe_dec (hat A) s with
                                 | left i => right (fun ni => ni i)
                                 | right ni => left ni end
    | step (recv start A) => in_dec eq_Probe_dec (bar A) s
    | step (send finish A) => in_dec eq_Probe_dec (hat A) s
    | step (recv finish A) => match in_dec eq_Probe_dec (bar A) s with
                                  | left i => right (fun ni => ni i)
                                  | right ni => left ni end
    | holds G => F G
    | holds2 G1 G2 => match F G1, F G2 with
                        | left r, left r' => left (conj r r')
                        | right nr, _ => right (fun r => nr (proj1 r))
                        | _, right nr => right (fun r => nr (proj2 r)) end
    | ll a0 a1 => match req_dec a0 s, req_dec a1 s with
                    | left r, left r' => left (conj r r')
                    | right nr, _ => right (fun r => nr (proj1 r))
                    | _, right nr => right (fun r => nr (proj2 r)) end 
    | _ => left I
  end.

Fixpoint Trace_to_NS {P P' a s s'} (t : Trace P a P') (r : req a s) (r' : ~ req a s')
                                      : Nonstable P s s' :=
  match t in Trace P a P' return req a s -> ~ req a s' -> Nonstable P s s'  with
    | CommB1_T A => fun r => fun r' => CommB1_NS _ _ _ r
                                       (match in_dec eq_Probe_dec _ _ with
                                          | left i => i
                                          | right ni => match r' ni with end end)
    | CommB2_T A => fun r => fun r' => CommB2_NS _ _ _ r r'
    | CommQ1_T A => fun r => fun r' => CommQ1_NS _ _ _ r r'
    | CommQ2_T A => fun r => fun r' => CommQ2_NS _ _ _ r
                                       (match in_dec eq_Probe_dec _ _ with
                                          | left i => i
                                          | right ni => match r' ni with end end)
    | Sequ_T _ _ _ _ t' => fun r => fun r' => Sequ_NS _ _ _ _ (Trace_to_NS t' r r')
    | Par_T _ _ _ _ a0 a1 t0 t1 => fun r => fun r' =>
                                   match req_dec a0 s' with
                                     | right nr0 => ParL_NS _ _ _ _ (Trace_to_NS t0 (proj1 r) nr0)
                                     | left r0 =>
                                   match req_dec a1 s' with
                                     | right nr1 => ParR_NS _ _ _ _ (Trace_to_NS t1 (proj2 r) nr1)
                                     | left r1 => match r' (conj r0 r1) with end end end
    | DetL_T (_, p0) _ => fun r => fun r' => DetL_NS _ _ _ _ (proj1 (eval_s1 _) r)
                                                               (fun x => r' (proj2 (eval_s1 _) x))
    | DetR_T _ (_, p1) => fun r => fun r' => DetR_NS _ _ _ _ (proj1 (eval_s1 _) r)
                                                               (fun x => r' (proj2 (eval_s1 _) x))
    | Det2_T (g0,_) (g1,_) => fun r => fun r' =>
                              match req_dec (holds g0) s' with
                                | right nr0 => DetL_NS _ _ _ _ (proj1 (eval_s1 _) (proj1 r))
                                                               (fun x => nr0 (proj2 (eval_s1 _) x))
                                | left r0 =>
                              match req_dec (holds g1) s' with
                                | right nr1 => DetR_NS _ _ _ _ (proj1 (eval_s1 _) (proj2 r))
                                                               (fun x => nr1 (proj2 (eval_s1 _) x))
                                | left r1 => match r' (conj r0 r1) with end end end
    | NDetL_T (_, p0) _ => fun r => fun r' => NDetL_NS _ _ _ _ (proj1 (eval_s1 _) r)
                                                               (fun x => r' (proj2 (eval_s1 _) x))
    | NDetR_T _ (_, p1) => fun r => fun r' => NDetR_NS _ _ _ _ (proj1 (eval_s1 _) r)
                                                               (fun x => r' (proj2 (eval_s1 _) x))
    | _ => fun _ => fun r' => match r' I with end
  end r r'.

Fixpoint Trace_to_NSE {P P' Q Q' a a' l} (t : Trace P a P') (t' : Trace Q a' Q')
                                         (r : req a l) (r' : req a' l)
                                         (r'' : ~ req a' (Label_to_newState a l))
                                         (f' : forall p, In p (pre a) -> In p l)
                                         (f : forall p, In p (down a) -> count_occ eq_Probe_dec l p = 1
                                                                      /\ count_occ eq_Probe_dec (pre a) p = 1)
                                            : {s : State & {R : Program & prod (SmallStep P l R s) (Nonstable Q l s)}} :=
  match t in Trace P a P'
  return req a l -> ~ req a' (Label_to_newState a l) -> (forall p, In p (pre a) -> In p l)
  -> (forall p, In p (down a) -> count_occ eq_Probe_dec l p = 1
                              /\ count_occ eq_Probe_dec (pre a) p = 1)
  -> {s : State & {R : Program & prod (SmallStep P l R s) (Nonstable Q l s)}} with
    | CommB1_T A => fun r => fun r'' => fun _ => fun f =>
                    existT _ _ (existT _ _ (CommB1_S A _ r, Trace_to_NS t' r' r''))
    | CommB2_T A => fun r => fun r'' => fun _ => fun f =>
                    existT _ _ (existT _ _ (CommB2_S A _ r (proj1 (f _ (in_eq _ _))), Trace_to_NS t' r' r''))
    | CommQ1_T A => fun r => fun r'' => fun _ => fun f =>
                    existT _ _ (existT _ _ (CommQ1_S A _ r, Trace_to_NS t' r' r''))
    | CommQ2_T A => fun r => fun r'' => fun _ => fun f =>
                    existT _ _ (existT _ _ (CommQ2_S A _ r (proj1 (f _ (in_eq _ _))), Trace_to_NS t' r' r''))
    | Sequ_T _ _ q _ t0 => fun r => fun r'' => fun f' => fun f =>
                           match Trace_to_NSE t0 t' r r' r'' f' f with
                           existT _ (existT _ (ss, ns)) => existT _ _ (existT _ _ (Sequ_S _ _ _ _ _ ss, ns)) end
    | Par_T p _ q _ a0 a1 t0 t1 => fun r => fun r'' => fun f' => fun f =>
                                   match req_dec a' (Label_to_newState a0 l) with
                                     | right nr0 => match Trace_to_NSE t0 t' (proj1 r) r' nr0
                                                          (fun p => fun i => f' p (in_or_app _ _ _ (or_introl i)))
                                                          (fun p => fun i => let X := (in_or_app _ _ _ (or_introl i))
                                                                             in (conj (proj1 (f p X))
                                                                                (proj2 (proj1 (in_count_occ_app_one (proj2 (f p X)))
                                                                                (down_to_pre i))))) with
                                                    existT _ (existT _ (ss, ns)) =>
                                                    existT _ _ (existT _ _ (ParL_S _ _ _ _ _ ss, ns)) end
                                     | left r0 =>
                                   match req_dec a' (Label_to_newState a1 l) with
                                     | right nr1 => match Trace_to_NSE t1 t' (proj2 r) r' nr1
                                                          (fun p => fun i => f' p (in_or_app _ _ _ (or_intror i)))
                                                          (fun p => fun i => let X := (in_or_app _ _ _ (or_intror i))
                                                                             in (conj (proj1 (f p X))
                                                                                (proj2 (proj2 (in_count_occ_app_one (proj2 (f p X)))
                                                                                (down_to_pre i))))) with
                                                    existT _ (existT _ (ss, ns)) =>
                                                    existT _ _ (existT _ _ (ParR_S _ _ _ _ _ ss, ns)) end
                                     | left r1 => match r'' (req_inv (proj2 r)
                                                                     (fun p => fun i => f' p (in_or_app _ _ _ (or_intror i)))
                                                                     (fun p => fun i => let X := (in_or_app _ _ _ (or_intror i))
                                                                                        in Label_to_Down a0 (f' p X)
                                                                     (fun ida0 => let Y := (in_or_app _ _ _ (or_introl ida0))
                                                                                  in proj1 (proj2 (in_count_occ_app_one (proj2 (f p Y)))
                                                                                  i) (down_to_pre ida0)))
                                                                     (fun p => fun i => let X := (in_or_app _ _ _ (or_intror i)) in
                                                                     (proj2 (proj2 (in_count_occ_app_one (proj2 (f p X)))
                                                                                   (down_to_pre i))))
                                                                      r' r1 r0) with end end end
    | _ => fun _ => fun r'' => fun _ => fun _ => match r'' r' with end
  end r r'' f' f.

Fixpoint Trace_to_ESorReq {P P' a l} (t : Trace P a P') (r : req a l) (f' : forall p, In p (pre a) -> In p l)
                            (f : forall p, In p (down a) -> count_occ eq_Probe_dec l p = 1
                                                         /\ count_occ eq_Probe_dec (pre a) p = 1)
                               : sum (ErrorStep P l) (req a (Label_to_newState a l)) :=
  match t as t in Trace P a P' return req a l -> (forall p, In p (pre a) -> In p l)
  -> (forall p, In p (down a) -> count_occ eq_Probe_dec l p = 1
                              /\ count_occ eq_Probe_dec (pre a) p = 1)
  -> sum (ErrorStep P l) (req a (Label_to_newState a l)) with
    | CommB1_T _ => fun r => fun _ => fun _ => inr (fun i => match in_inv i with
                                                      | or_introl e => proj1 neq_Probe e
                                                      | or_intror i' => r i' end)
    | CommB2_T _ => fun r => fun _ => fun _ => inr (in_remove_neq (proj2 neq_Probe) r)
    | CommQ1_T _ => fun r => fun _ => fun _ => inr (in_cons _ _ _ r)
    | CommQ2_T _ => fun r => fun _ => fun _ => inr (fun i => r (in_remove_weak i))
    | Sequ_T _ _ _ _ t' => fun r => fun f' => fun f =>
                           match Trace_to_ESorReq t' r f' f with
                             | inl es => inl (Sequ_E _ _ _ es)
                             | inr r' => inr r' end
    | Par_T _ _ _ _ a0 a1 t0 t1 => fun r => fun f' => fun f =>
                                   match Trace_to_ESorReq t0 (proj1 r)
                                         (fun p => fun i => f' p (in_or_app _ _ _ (or_introl i)))
                                         (fun p => fun i => let X := (in_or_app _ _ _ (or_introl i))
                                                            in (conj (proj1 (f p X))
                                                               (proj2 (proj1 (in_count_occ_app_one (proj2 (f p X)))
                                                               (down_to_pre i))))) with
                                     | inl es => inl (ParL_E _ _ _ es)
                                     | inr r00 =>
                                   match Trace_to_ESorReq t1 (proj2 r)
                                         (fun p => fun i => f' p (in_or_app _ _ _ (or_intror i)))
                                         (fun p => fun i => let X := (in_or_app _ _ _ (or_intror i))
                                                            in (conj (proj1 (f p X))
                                                               (proj2 (proj2 (in_count_occ_app_one (proj2 (f p X)))
                                                               (down_to_pre i))))) with
                                     | inl es' => inl (ParR_E _ _ _ es')
                                     | inr r11 =>
                                   match req_dec a1 (Label_to_newState a0 l) with
                                     | right nr10 => match Trace_to_NSE t0 t1 (proj1 r) (proj2 r) nr10
                                                           (fun p => fun i => f' p (in_or_app _ _ _ (or_introl i)))
                                         (fun p => fun i => let X := (in_or_app _ _ _ (or_introl i))
                                                            in (conj (proj1 (f p X))
                                                               (proj2 (proj1 (in_count_occ_app_one (proj2 (f p X)))
                                                               (down_to_pre i))))) with
                                                     existT _ (existT _ (ss, ns)) => inl (ParL_NSE _ _ _ _ _ ss ns) end
                                     | left r10 =>
                                   match req_dec a0 (Label_to_newState a1 l) with
                                     | right nr01 => match Trace_to_NSE t1 t0 (proj2 r) (proj1 r) nr01
                                                           (fun p => fun i => f' p (in_or_app _ _ _ (or_intror i)))
                                         (fun p => fun i => let X := (in_or_app _ _ _ (or_intror i))
                                                            in (conj (proj1 (f p X))
                                                               (proj2 (proj2 (in_count_occ_app_one (proj2 (f p X)))
                                                               (down_to_pre i))))) with
                                                     existT _ (existT _ (ss, ns)) => inl (ParR_NSE _ _ _ _ _ ss ns) end
                                     | left r01 => inr (conj (req_inv (proj2 r)
                                                                      (fun p => fun i => f' p (in_or_app _ _ _ (or_intror i)))
                                                                      (fun p => fun i => let X := (in_or_app _ _ _ (or_intror i))
                                                                                        in Label_to_Down a0 (f' p X)
                                                                      (fun ida0 => let Y := (in_or_app _ _ _ (or_introl ida0))
                                                                                   in proj1 (proj2 (in_count_occ_app_one (proj2 (f p Y)))
                                                                                   i) (down_to_pre ida0)))
                                                                      (fun p => fun i => let X := (in_or_app _ _ _ (or_intror i)) in
                                                                      (proj2 (proj2 (in_count_occ_app_one (proj2 (f p X)))
                                                                                    (down_to_pre i))))
                                                                      (proj1 r) r01 r00)
                                                             (req_inv (proj2 r)
                                                                      (fun p => fun i => f' p (in_or_app _ _ _ (or_intror i)))
                                                                      (fun p => fun i => let X := (in_or_app _ _ _ (or_intror i))
                                                                                        in Label_to_Down a0 (f' p X)
                                                                      (fun ida0 => let Y := (in_or_app _ _ _ (or_introl ida0))
                                                                                   in proj1 (proj2 (in_count_occ_app_one (proj2 (f p Y)))
                                                                                   i) (down_to_pre ida0)))
                                                                      (fun p => fun i => let X := (in_or_app _ _ _ (or_intror i)) in
                                                                      (proj2 (proj2 (in_count_occ_app_one (proj2 (f p X)))
                                                                                    (down_to_pre i))))
                                                                      (proj2 r) r11 r10)) end end end end
    | DetL_T _ _ => fun r => fun _ => fun _ => inr r
    | DetR_T _ _ => fun r => fun _ => fun _ => inr r
    | Det2_T _ _ => fun r => fun _ => fun _ => inr r
    | NDetL_T _ _ => fun r => fun _ => fun _ => inr r
    | NDetR_T _ _ => fun r => fun _ => fun _ => inr r
    | _ => fun _ => fun _ => fun _ => inr I
  end r f' f.

Fixpoint Label_to_DownCount {p l} (a : Label) (i : ~ In p (down a)) (i' : ~ In p (post a))
                            : count_occ eq_Probe_dec (Label_to_newState a l) p = count_occ eq_Probe_dec l p :=
  match a as a return ~ In p (down a) -> ~ In p (post a)
  -> count_occ eq_Probe_dec (Label_to_newState a l) p = count_occ eq_Probe_dec l p with
    | step (send start A) => fun i => fun i' => match eq_Probe_dec (bar A) p with
                                          | left e => match i' (eq_ind_r (fun x => In x _) (in_eq _ _) (eq_sym e)) with end
                                          | right ne => count_occ_cons_neq _ _ ne end
    | step (recv start A) => fun i => fun i' => match eq_Probe_dec (hat A) p with
                                          | left e => match i' (eq_ind_r (fun x => In x _) (in_eq _ _) (eq_sym e)) with end
                                          | right ne => count_occ_cons_neq _ _ ne end
    | step (send finish A) => fun i => fun i' => match eq_Probe_dec p (bar A) with
                                           | left e => match i (eq_ind_r (fun x => In x _) (in_eq _ _) e) with end
                                           | right ne => count_occ_remove_neq ne end
    | step (recv finish A) => fun i => fun i' => match eq_Probe_dec p (hat A) with
                                           | left e => match i (eq_ind_r (fun x => In x _) (in_eq _ _) e) with end
                                           | right ne => count_occ_remove_neq ne end
    | ll a0 a1 => fun i => fun i' => eq_trans (Label_to_DownCount a1 (fun x => i (in_or_app _ _ _ (or_intror x)))
                                                                 (fun x => i' (in_or_app _ _ _ (or_intror x))))
                                              (Label_to_DownCount a0 (fun x => i (in_or_app _ _ _ (or_introl x)))
                                                                 (fun x => i' (in_or_app _ _ _ (or_introl x))))
    | _ => fun _ => fun _ => eq_refl _
  end i i'.

Fixpoint Label_to_Count {s l p n m} (a : Label) (r : req a s)
                        (f : forall p, In p (down a) -> count_occ eq_Probe_dec l p = 1
                                                     /\ count_occ eq_Probe_dec (pre a) p = 1)
                        (nm : count_occ eq_Probe_dec l p + n = count_occ eq_Probe_dec (pre a) p + m)
                            : count_occ eq_Probe_dec (Label_to_newState a l) p + n = count_occ eq_Probe_dec (post a) p + m :=
  match a as a return req a s
  -> (forall p, In p (down a) -> count_occ eq_Probe_dec l p = 1
                              /\ count_occ eq_Probe_dec (pre a) p = 1)
  -> count_occ eq_Probe_dec l p + n = count_occ eq_Probe_dec (pre a) p + m
  -> count_occ eq_Probe_dec (Label_to_newState a l) p + n = count_occ eq_Probe_dec (post a) p + m with
    | step (send start A) => fun _ => fun _ => fun nm =>
                               match eq_Probe_dec (bar A) p with
                                 | left e => count_cons_eq nm (count_occ_cons_eq eq_Probe_dec l e)
                                                              (eq_sym (count_occ_cons_eq eq_Probe_dec nil e))
                                 | right ne => count_trans_neq nm (count_occ_cons_neq eq_Probe_dec l ne)
                                                                  (eq_sym (count_occ_cons_neq eq_Probe_dec nil ne)) end
    | step (recv start A) => fun _ => fun _ => fun nm =>
                               match eq_Probe_dec (hat A) p with
                                 | left e => count_cons_eq nm (count_occ_cons_eq eq_Probe_dec l e)
                                                              (eq_sym (count_occ_cons_eq eq_Probe_dec nil e))
                                 | right ne => count_trans_neq nm (count_occ_cons_neq eq_Probe_dec l ne)
                                                                  (eq_sym (count_occ_cons_neq eq_Probe_dec nil ne)) end
    | step (send finish A) => fun _ => fun f => fun nm =>
                                match eq_Probe_dec p (bar A) with
                                  | left e => count_remove_eq nm (proj1 count_occ_not_In (eq_ind_r (fun x => ~ In x _) (remove_In _ _ _) e))
                                                                 (proj1 (f p (eq_ind_r (fun x => In x _) (in_eq _ _) e)))
                                                                 (eq_ind_r (fun x => count_occ _ _ x = _) count_sing_one e)
                                                                 (proj1 count_occ_not_In (in_nil (a := p)))
                                  | right ne => count_trans_neq nm (count_occ_remove_neq ne)
                                                                   (count_occ_cons_neq eq_Probe_dec nil (fun e => ne (eq_sym e))) end
    | step (recv finish A) => fun _ => fun f => fun nm =>
                                match eq_Probe_dec p (hat A) with
                                  | left e => count_remove_eq nm (proj1 count_occ_not_In (eq_ind_r (fun x => ~ In x _) (remove_In _ _ _) e))
                                                                 (proj1 (f p (eq_ind_r (fun x => In x _) (in_eq _ _) e)))
                                                                 (eq_ind_r (fun x => count_occ _ _ x = _) count_sing_one e)
                                                                 (proj1 count_occ_not_In (in_nil (a := p)))
                                  | right ne => count_trans_neq nm (count_occ_remove_neq ne)
                                                                   (count_occ_cons_neq eq_Probe_dec nil (fun e => ne (eq_sym e))) end
    | ll a0 a1 => fun r => fun f => fun nm =>
                  let c0 := Label_to_Count (l := l) a0 (proj1 r)
                            (fun p => fun i => let X := (in_or_app _ _ _ (or_introl i))
                                               in (conj (proj1 (f p X))
                                                  (proj2 (proj1 (in_count_occ_app_one (proj2 (f p X)))
                                                  (down_to_pre i)))))
                            (eq_sym (proj1 (eq_plus_assoc (or_introl count_occ_app)) (eq_sym nm)))
                  in let c1 := Label_to_Count (l := Label_to_newState a0 l) a1 (proj2 r)
                               (fun p => fun i => let X := (in_or_app _ _ _ (or_intror i))
                                                  in (conj (eq_trans (Label_to_DownCount a0
                                                     (fun id0 => proj1 (proj1 (in_count_occ_app_one (proj2 (f p X)))
                                                     (down_to_pre id0)) (down_to_pre i))
                                                     (fun ip0 => proj1 (proj1 (in_count_occ_app_one (proj2 (f p X)))
                                                     (Labels_to_pre (conj (proj2 r) (proj1 r)) i ip0)) (down_to_pre i)))
                                                     (proj1 (f p X))) (proj2 (proj2 (in_count_occ_app_one
                                                     (proj2 (f p X))) (down_to_pre i)))))
                               (plus_swap c0)
                  in (eq_sym (proj2 (eq_plus_assoc (or_intror count_occ_app)) (eq_sym c1)))
    | _ => fun _ => fun _ => fun nm => nm
  end r f nm.

Fixpoint Label_to_PrePost (a : Label) (c : compliant a)
                          (f : forall p, In p (down a) -> count_occ eq_Probe_dec (pre a) p = 1)
                             : eq_State (Label_to_newState a (pre a)) (post a) :=
  fun p => eq_trans (plus_n_O _) (eq_trans
           (Label_to_Count a c (fun p => fun i => conj (f p i) (f p i))
           (eq_ind_r (fun x => x = x) (eq_refl _) (plus_n_O _)))
           (eq_sym (plus_n_O _))).

Definition Label_dec (a : Label) : (forall p, In p (down a) -> count_occ eq_Probe_dec (pre a) p = 1)
                                 + {p : Probe | In p (down a) /\ count_occ eq_Probe_dec (pre a) p > 1} :=
  (fix F l := match l as l
              return (forall p, In p l -> In p (down a))
              -> (forall p, In p l -> count_occ eq_Probe_dec (pre a) p = 1)
               + {p : Probe | In p l /\ count_occ eq_Probe_dec (pre a) p > 1} with
                | nil => fun _ => inl (fun _ => fun i => match in_nil i with end)
                | x :: l' => fun f =>
                             match in_once_or_more (down_to_pre (f x (in_eq _ _))) with
                               | right gt1 => inr (exist _ x (conj (in_eq _ _) gt1))
                               | left eq1 =>
                             match F l' (fun p => fun i => f p (in_cons _ _ _ i)) with
                               | inr (exist x (conj i' gt1')) => inr (exist _ x (conj (in_cons _ _ _ i') gt1'))
                               | inl eq1' => inl (fun p => fun i =>
                                                  match in_inv (a := x) (b := p) i with
                                                    | or_introl e => eq_ind_r (fun x => count_occ _ _ x = 1) eq1 (eq_sym e)
                                                    | or_intror i' => eq1' p i' end) end end
              end) (down a) (fun _ => fun i => i).

Definition TraceNotStable_to_ES {P P' a} (t : Trace P a P') (c : compliant a)
                                         (ns : ~ stable a) : ErrorStep P (pre a) :=
  match Label_dec a with
    | inl f => match Trace_to_ESorReq t c (fun p => fun i => i) (fun p => fun i => conj (f p i) (f p i)) with
                 | inl ES => ES
                 | inr r => match ns (req_eq_set r (eq_State_to_eq_set (Label_to_PrePost a c f))) with end end
    | inr p => match p with exist _ ic => TraceDownCount_to_ES t c ic end end.